# -*- coding: utf-8 -*-
"""A simple command line application to download youtube videos."""
from __future__ import absolute_import
from __future__ import print_function

import argparse
import datetime as dt
import gzip
import json
import logging
import os
import sys

from pytube import __version__
from pytube import request
from pytube import YouTube


logger = logging.getLogger(__name__)


def main():
    """Command line application to download youtube videos."""
    parser = argparse.ArgumentParser(description=main.__doc__)
    parser.add_argument('url', help='The YouTube /watch url', nargs='?')
    parser.add_argument(
        '--version', action='version',
        version='%(prog)s ' + __version__,
    )
    parser.add_argument(
        '--itag', type=int, help=(
            'The itag for the desired stream'
        ),
    )
    parser.add_argument(
        '-l', '--list', action='store_true', help=(
            'The list option causes pytube cli to return a list of streams '
            'available to download'
        ),
    )
    parser.add_argument(
        '-v', '--verbose', action='count', default=0, dest='verbosity',
        help='Verbosity level',
    )
    parser.add_argument(
        '--build-playback-report', action='store_true', help=(
            'Save the html and js to disk'
        ),
    )

    args = parser.parse_args()
    logging.getLogger().setLevel(max(3 - args.verbosity, 0) * 10)

    if not args.url:
        parser.print_help()
        sys.exit(1)

    if args.list:
        display_streams(args.url)

    elif args.build_playback_report:
        build_playback_report(args.url)

    elif args.itag:
        download(args.url, args.itag)


def build_playback_report(url):
    """Serialize the request data to json for offline debugging.

    :param str url:
        A valid YouTube watch URL.
    """
    yt = YouTube(url)
    ts = int(dt.datetime.utcnow().timestamp())
    fp = os.path.join(
        os.getcwd(),
        'yt-video-{yt.video_id}-{ts}.json.tar.gz'.format(yt=yt, ts=ts),
    )
    js, watch_html, vid_info = request.get(urls=[
        yt.js_url,
        yt.watch_url,
        yt.vid_info_url,
    ])
    with gzip.open(fp, 'wb') as fh:
        fh.write(
            json.dumps({
                'url': url,
                'js': js,
                'watch_html': watch_html,
                'video_info': vid_info,
            })
            .encode('utf8'),
        )


def get_terminal_size():
    """Return the terminal size in rows and columns."""
    rows, columns = os.popen('stty size', 'r').read().split()
    return int(rows), int(columns)


def display_progress_bar(bytes_received, filesize, ch='█', scale=0.55):
    """Display a simple, pretty progress bar.

    Example:
    ~~~~~~~~
    PSY - GANGNAM STYLE(강남스타일) MV.mp4
    ↳ |███████████████████████████████████████| 100.0%

    :param int bytes_received:
        The delta between the total file size (bytes) and bytes already
        written to disk.
    :param int filesize:
        File size of the media stream in bytes.
    :param ch str:
        Character to use for presenting progress segment.
    :param float scale:
        Scale multipler to reduce progress bar size.

    """
    _, columns = get_terminal_size()
    max_width = int(columns * scale)

    filled = int(round(max_width * bytes_received / float(filesize)))
    remaining = max_width - filled
    bar = ch * filled + ' ' * remaining
    percent = round(100.0 * bytes_received / float(filesize), 1)
    text = ' ↳ |{bar}| {percent}%\r'.format(bar=bar, percent=percent)
    sys.stdout.write(text)
    sys.stdout.flush()


def on_progress(stream, file_handle, bytes_remaining):
    """On download progress callback function.

    :param object stream:
        An instance of :class:`Stream <Stream>` being downloaded.
    :param :class:`_io.BufferedWriter <BufferedWriter>` file_handle:
        The file handle where the media is being written to.
    :param int bytes_remaining:
        How many bytes have been downloaded.

    """
    filesize = stream.filesize
    bytes_received = filesize - bytes_remaining
    display_progress_bar(bytes_received, filesize)


def download(url, itag):
    """Start downloading a YouTube video.

    :param str url:
        A valid YouTube watch URL.
    :param str itag:
        YouTube format identifier code.

    """
    # TODO(nficano): allow download target to be specified
    # TODO(nficano): allow dash itags to be selected
    yt = YouTube(url, on_progress_callback=on_progress)
    stream = yt.streams.get(itag)
    print('\n{fn} | {fs} bytes'.format(
        fn=stream.default_filename,
        fs=stream.filesize,
    ))
    try:
        stream.download()
        sys.stdout.write('\n')
    except KeyboardInterrupt:
        sys.exit()


def display_streams(url):
    """Probe YouTube video and lists its available formats.

    :param str url:
        A valid YouTube watch URL.

    """
    yt = YouTube(url)
    for stream in yt.streams.all():
        print(stream)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Applies in-place data mutations."""
from __future__ import absolute_import

import logging
import pprint

from pytube import cipher
from pytube.compat import parse_qsl
from pytube.compat import unquote


logger = logging.getLogger(__name__)


def apply_signature(config_args, fmt, js):
    """Apply the decrypted signature to the stream manifest.

    :param dict config_args:
        Details of the media streams available.
    :param str fmt:
        Key in stream manifests ("player_config") containing progressive
        download or adaptive streams (e.g.: "url_encoded_fmt_stream_map" or
        "adaptive_fmts").
    :param str js:
        The contents of the base.js asset file.

    """
    stream_manifest = config_args[fmt]
    for i, stream in enumerate(stream_manifest):
        url = stream['url']

        if 'signature=' in url:
            # For certain videos, YouTube will just provide them pre-signed, in
            # which case there's no real magic to download them and we can skip
            # the whole signature decrambling entirely.
            continue

        signature = cipher.get_signature(js, stream['s'])

        logger.debug(
            'finished descrambling signature for itag=%s\n%s',
            stream['itag'], pprint.pformat(
                {
                    's': stream['s'],
                    'signature': signature,
                }, indent=2,
            ),
        )
        stream_manifest[i]['url'] = url + '&signature=' + signature


def apply_descrambler(stream_data, key):
    """Apply various in-place transforms to YouTube's media stream data.

    :param dict dct:
        Dictionary containing query string encoded values.
    :param str key:
        Name of the key in dictionary.

    """
    stream_data[key] = [
        {k: unquote(v) for k, v in parse_qsl(i)}
        for i in stream_data[key].split(',')
    ]
    logger.debug('applying descrambler\n%s', pprint.pformat(stream_data[key]))



# -*- coding: utf-8 -*-
"""Reusable dependency injected testing components."""
import json
import os

import pytest

from pytube import YouTube


@pytest.fixture
def gangnam_style():
    """Youtube instance initialized with video id 9bZkp7q19f0."""
    cur_dir = os.path.dirname(os.path.realpath(__file__))
    fp = os.path.join(cur_dir, 'mocks', 'yt-video-9bZkp7q19f0.json')
    video = None
    with open(fp, 'r') as fh:
        video = json.loads(fh.read())
    yt = YouTube(
        url='https://www.youtube.com/watch?v=9bZkp7q19f0',
        defer_prefetch_init=True,
    )
    yt.watch_html = video['watch_html']
    yt.js = video['js']
    yt.vid_info = video['video_info']
    yt.init()
    return yt



# -*- coding: utf-8 -*-
"""
This module countains all logic necessary to decipher the signature.

YouTube's strategy to restrict downloading videos is to send a ciphered version
of the signature to the client, along with the decryption algorithm obfuscated
in JavaScript. For the clients to play the videos, JavaScript must take the
ciphered version, pass it through a series of "transform functions," and then
signs the media URL with the output. On the backend, they verify that the
signature sent in the GET parameters is valid, and then returns the content
(video/audio stream) or 403 unauthorized accordingly.

This module is responsible for (1) finding and extracting those "transform
functions" (2) maps them to Python equivalents and (3) taking the ciphered
signature and decoding it.

"""
from __future__ import absolute_import

import logging
import pprint
import re
from itertools import chain

from pytube.exceptions import RegexMatchError
from pytube.helpers import memoize
from pytube.helpers import regex_search


logger = logging.getLogger(__name__)


def get_initial_function_name(js):
    """Extract the name of the function responsible for computing the signature.

    :param str js:
        The contents of the base.js asset file.

    """
    # c&&d.set("signature", EE(c));
    pattern = r'"signature",\s?([a-zA-Z0-9$]+)\('
    logger.debug('finding initial function name')
    return regex_search(pattern, js, group=1)


def get_transform_plan(js):
    """Extract the "transform plan".

    The "transform plan" is the functions the ciphered signature is passed
    through to obtain the actual signature.

    :param str js:
        The contents of the base.js asset file.

    Sample Output:
    ~~~~~~~~~~~~~~
    ['DE.AJ(a,15)',
     'DE.VR(a,3)',
     'DE.AJ(a,51)',
     'DE.VR(a,3)',
     'DE.kT(a,51)',
     'DE.kT(a,8)',
     'DE.VR(a,3)',
     'DE.kT(a,21)']

    """
    name = re.escape(get_initial_function_name(js))
    pattern = r'%s=function\(\w\){[a-z=\.\(\"\)]*;(.*);(?:.+)}' % name
    logger.debug('getting transform plan')
    return regex_search(pattern, js, group=1).split(';')


def get_transform_object(js, var):
    """Extract the "transform object".

    The "transform object" contains the function definitions referenced in the
    "transform plan". The ``var`` argument is the obfuscated variable name
    which contains these functions, for example, given the function call
    ``DE.AJ(a,15)`` returned by the transform plan, "DE" would be the var.

    :param str js:
        The contents of the base.js asset file.
    :param str var:
        The obfuscated variable name that stores an object with all functions
        that descrambles the signature.

    Sample Output:
    ~~~~~~~~~~~~~~
    ['AJ:function(a){a.reverse()}',
     'VR:function(a,b){a.splice(0,b)}',
     'kT:function(a,b){var c=a[0];a[0]=a[b%a.length];a[b]=c}']

    """
    pattern = r'var %s={(.*?)};' % re.escape(var)
    logger.debug('getting transform object')
    return (
        regex_search(pattern, js, group=1, flags=re.DOTALL)
        .replace('\n', ' ')
        .split(', ')
    )


@memoize
def get_transform_map(js, var):
    """Build a transform function lookup.

    Build a lookup table of obfuscated JavaScript function names to the
    Python equivalents.

    :param str js:
        The contents of the base.js asset file.
    :param str var:
        The obfuscated variable name that stores an object with all functions
        that descrambles the signature.

    """
    transform_object = get_transform_object(js, var)
    mapper = {}
    for obj in transform_object:
        # AJ:function(a){a.reverse()} => AJ, function(a){a.reverse()}
        name, function = obj.split(':', 1)
        fn = map_functions(function)
        mapper[name] = fn
    return mapper


def reverse(arr, b):
    """Reverse elements in a list.

    This function is equivalent to: function(a, b) { a.reverse() }.

    This method takes an unused ``b`` variable as their transform functions
    universally sent two arguments.

    Example usage:
    ~~~~~~~~~~~~~~
    >>> reverse([1, 2, 3, 4])
    [4, 3, 2, 1]
    """
    return arr[::-1]


def splice(arr, b):
    """Add/remove items to/from a list.

    This function is equivalent to: function(a, b) { a.splice(0, b) }.

    Example usage:
    ~~~~~~~~~~~~~~
    >>> splice([1, 2, 3, 4], 2)
    [1, 2]
    """
    return arr[:b] + arr[b * 2:]


def swap(arr, b):
    """Swap positions at b modulus the list length.

    This function is equivalent to:
    function(a, b) { var c=a[0];a[0]=a[b%a.length];a[b]=c }.

    Example usage:
    ~~~~~~~~~~~~~~
    >>> swap([1, 2, 3, 4], 2)
    [3, 2, 1, 4]
    """
    r = b % len(arr)
    return list(chain([arr[r]], arr[1:r], [arr[0]], arr[r + 1:]))


def map_functions(js_func):
    """For a given JavaScript transform function, return the Python equivalent.

    :param str js_func:
        The JavaScript version of the transform function.

    """
    mapper = (
        # function(a){a.reverse()}
        ('{\w\.reverse\(\)}', reverse),
        # function(a,b){a.splice(0,b)}
        ('{\w\.splice\(0,\w\)}', splice),
        # function(a,b){var c=a[0];a[0]=a[b%a.length];a[b]=c}
        ('{var\s\w=\w\[0\];\w\[0\]=\w\[\w\%\w.length\];\w\[\w\]=\w}', swap),
    )

    for pattern, fn in mapper:
        if re.search(pattern, js_func):
            return fn
    raise RegexMatchError(
        'could not find python equivalent function for: ',
        js_func,
    )


def parse_function(js_func):
    """Parse the Javascript transform function.

    Break a JavaScript transform function down into a two element tuple
    containing the function name and some integer-based argument.

    Sample Input:
    ~~~~~~~~~~~~~
    DE.AJ(a,15)

    Sample Output:
    ~~~~~~~~~~~~~~
    ('AJ', 15)

    :param str js_func:
        The JavaScript version of the transform function.

    """
    pattern = r'\w+\.(\w+)\(\w,(\d+)\)'
    logger.debug('parsing transform function')
    return regex_search(pattern, js_func, groups=True)


@memoize
def get_signature(js, ciphered_signature):
    """Decipher the signature.

    Taking the ciphered signature, applies the transform functions and
    returns the decrypted version.

    :param str js:
        The contents of the base.js asset file.
    :param str ciphered_signature:
        The ciphered signature sent in the ``player_config``.

    """
    tplan = get_transform_plan(js)
    # DE.AJ(a,15) => DE, AJ(a,15)
    var, _ = tplan[0].split('.')
    tmap = get_transform_map(js, var)
    signature = [s for s in ciphered_signature]

    for js_func in tplan:
        name, argument = parse_function(js_func)
        signature = tmap[name](signature, int(argument))
        logger.debug(
            'applied transform function\n%s', pprint.pformat(
                {
                    'output': ''.join(signature),
                    'js_function': name,
                    'argument': int(argument),
                    'function': tmap[name],
                }, indent=2,
            ),
        )
    return ''.join(signature)



# -*- coding: utf-8 -*-
"""Implements a simple wrapper around urlopen."""
import multiprocessing

from pytube.compat import urlopen


def get(
    url=None, urls=[], processes=2, headers=False,
    streaming=False, chunk_size=8 * 1024,
):
    """Send an http GET request.

    :param str url:
        The URL to perform the GET request for.
    :param list urls:
        List of URLs to perform the GET request for concurrently.
    :param int processes:
        How many worker processes to start.
    :param bool headers:
        Only return the http headers.
    :param bool streaming:
        Returns the response body in chunks via a generator.
    :param int chunk_size:
        The size in bytes of each chunk.
    """
    if urls:
        pool = multiprocessing.Pool(processes=processes)
        return pool.map(get, urls)
    else:
        response = urlopen(url)
        if streaming:
            return stream_response(response, chunk_size)
        elif headers:
            return dict(response.info().items())
        return (
            response
            .read()
            .decode('utf-8')
        )


def stream_response(response, chunk_size=8 * 1024):
    """Read the :module:`urlopen` response in chunks."""
    while True:
        buf = response.read(chunk_size)
        if not buf:
            break
        yield buf



# -*- coding: utf-8 -*-
"""Library specific exception definitions."""
import socket
import sys
from traceback import format_tb

from pytube.compat import URLError


class PytubeError(Exception):
    """Base pytube exception that all others inherent.

    This is done to not pollute the built-in exceptions, which *could* result
    in unintended errors being unexpectedly and incorrectly handled within
    implementers code.
    """


class ExtractError(PytubeError):
    """Data extraction based exception."""

    def __init__(self, msg, tb=None, expected=False, video_id=None):
        """Construct an instance of a :class:`ExtractError <ExtractError>`.

        :param str msg:
            User defined error message.
        :param list tb:
            Stack trace leading up to the exception.
        :param bool expected:
            Whether the exception being raised requires filing a bug report.
        :param str video_id:
            A YouTube video identifer.
        """
        # get information about the most recent exception caught by an except
        # clause
        exception_type, _, _ = sys.exc_info()

        if exception_type in (URLError, socket.timeout):
            expected = True

        if video_id is not None:
            msg = '{video_id}: {msg}'.format(video_id=video_id, msg=msg)

        if not expected:
            # TODO(nficano): bug report
            pass

        super(ExtractError, self).__init__(msg)

        self.traceback = tb
        self.exc_info = sys.exc_info()
        self.video_id = video_id

    def format_traceback(self):
        """Pretty-print the traceback."""
        if self.traceback:
            return ''.join(format_tb(self.traceback))


class RegexMatchError(ExtractError):
    """Regex pattern did not return any matches."""
